#!/usr/bin/perl

use strict;
use warnings;
use Time::Seconds;


require 'shelf-lib.pl';

BEGIN { push(@INC, ".."); };
use WebminCore;
our(%text, %config, %gconfig, %store, $host, %pool_color);

my $refresh = undef;

&init_config();
&ReadParse();

sub shelf_footer
{

	print &ui_print_footer("", $text{'shelf_list_return'});
}


# Prints shelf info
sub shelf_show
{
	my ($addr) = @_;
	my $shelf = \%{$store{'shelf'}{$addr}};
	my $shelf_name = $shelf->{'name'};
	my $status;
	my $targets = 0;
	my $size = 0;
	my $uptime = $shelf->{'uptime'};
	my $color = undef;
	my $t = Time::Seconds->new($uptime);
	my @row;
	my %sinfo;

	if (!defined($shelf_name) || ($shelf_name eq "")) {
		$shelf_name = "unknown";
	}

#	print "<pre>".Dumper($shelf)."</pre>";
	$sinfo{'uptime'} = $uptime;
	shelf_info(\%sinfo, $shelf->{'target'});
#	print "<pre>".Dumper(\%sinfo)."</pre>";
#	if (defined($shelf->{'target'}) && (scalar($shelf->{'target'}) > 0)) {
#		# here we have a target from an IP address
#		my $bgc;
#		if (defined($shelf->{'vdev'}) &&
#			(scalar($shelf->{'vdev'}) > 0)) {
#			# a list of vdevs from the shelf
#			$status = $text{"shelf_status_online"};
##			$color = "bgcolor=\"#00D000\"";
#			$size = 0;
#			if (has_access("shelf_detail")) {
#				my $url = "?cmd=detail&shelf=$shelf_name";
#				push(@row, "<a href=\"$url\">".$shelf_name."</a>");
#			} else {
#				push(@row, $shelf_name);
#			}
##			print "<pre>".Dumper($shelf->{'pool'})."</pre>";
#		} else {
#			# target only
#			$status = $text{"shelf_status_present"};
#			$bgc = "yellow";
##			$color = "bgcolor=\"yellow\"";
#			push(@row, $shelf_name);
#		}
#		my $stc = 0;
#		foreach my $p (sort keys %{$shelf->{'pool'}}) {
#			my $pool = \%{$shelf->{'pool'}{$p}};
##			print "<pre>".Dumper($pool)."</pre>";
#			if (($pool->{'status'} =~ /online/i) && ($stc == 0)) {
#				$stc = 1;
#			} elsif (($pool->{'status'} =~ /offline/i) && ($stc < 2)) {
#				$stc = 2;
#			} elsif (($pool->{'status'} =~ /degraded/i) && ($stc < 3)) {
#				$stc = 3;
#			} elsif (($pool->{'status'} =~ /unavail/i) && ($stc < 4)) {
#				$stc = 4;
#				last;
#			}
#		}
#		if (($stc == 0) || ($stc == 1)) {
#			if (!defined($bgc)) {
#				$bgc = "#00D000";
#			}
#			$status .= ", ONLINE";
#		} elsif ($stc == 2) {
#			if (!defined($bgc)) {
#				$bgc = "yellow";
#			}
#			$status .= ", OFFLINE";
#		} elsif ($stc == 3) {
#			if (!defined($bgc)) {
#				$bgc = "yellow";
#			}
#			$status .= ", DEGRADED";
#		} else {
#			$bgc = "red";
#			$status .= ", UNAVAIL";
#		}
#		$color = "bgcolor=\"$bgc\"";
#	} else {
#		$status = $text{"shelf_status_absent"};
#		$color = "bgcolor=\"#E00000\"";
#		push(@row, $shelf_name);
#	}
	if (has_access("shelf_detail") && defined($uptime)) {
		my $url = "?cmd=detail&shelf=$shelf_name";
		push(@row, "<a href=\"$url\">".$shelf_name."</a>");
	} else {
		push(@row, $shelf_name);
	}
	push(@row, $sinfo{'status'});
#	$predict = smart_prediction($store{'addr'}{$addr}{'smart'});
#	if ($predict < 1.0) {
#		# OK
#		$str = "OK";
#		$colorp = "bgcolor=\"#00ff00\""
#	} elsif ($predict < 10.0) {
#		# warning
#		$str = "Warning";
#		$colorp = "bgcolor=\"yellow\""
#	} elsif ($predict < 50.0) {
#		# critical
#		$str = "Critical";
#		$colorp = "bgcolor=\"red\""
#	} else {
#		# catastrophe ;-)
#		$str = "Amen ;-)";
#		$colorp = "bgcolor=\"#cc9966\""
#	}
#	push(@row, $str);
	if (has_access("shelf_addr")) {
		push(@row, $addr);
	}
	if (has_access("shelf_size")) {
		my $str;
		if ($size < 0) {
			$str = "?";
		} else {
			$size = shelf_target_size($addr);
			$str = nice_size($size);
		}
		push(@row, $str);
	}
	if (has_access("shelf_uptime")) {
		push(@row, $t->pretty);
	}
	if (has_access("lun_manage")) {
		# FIXME: in progress
		# проверить если в пуле, то сделать online/offline
		# в противном случае написать, что не используется
		push(@row, "offline");
	}
	if (defined($sinfo{'color'})) {
		$color = "bgcolor=\"".$sinfo{'color'}."\"";
	}
	print &ui_columns_row(\@row, [undef, $color]);
}


sub shelves_show
{
	my $addr;
	my @hdr = ($text{"shelf_name"}, $text{"shelf_status"});
	my @alist = ();

	if (has_access("shelf_addr")) {
		push(@hdr, $text{"shelf_addr"});
	}
	if (has_access("shelf_size")) {
		push(@hdr, $text{"shelf_size"});
	}
	if (has_access("shelf_uptime")) {
		push(@hdr, $text{"shelf_uptime"});
	}
	if (has_access("lun_manage")) {
		push(@hdr, $text{"lun_ctl"});
	}

	print &ui_table_start($text{"shelf_list"}, "width=100%", scalar @hdr);

	print &ui_columns_start(\@hdr);
	foreach my $tl (keys %{$store{'config'}}) {
		if (($tl =~ /^pool_.*_targets/i) && !($tl =~ /^pool_.*_targets_min/i)) {
			push(@alist, @{$store{'config'}{$tl}});
		}
	}
	my %uhash = map { $_ => 1} @alist;
	@alist = keys %uhash;
	foreach $addr (sort @alist) {
		$addr =~ s/:.*//;
		shelf_show($addr);
	}
	print &ui_columns_end();

#	print "<tr>";
#	print "<td>";
#	print &ui_form_start('index.cgi', 'post');
#	print &ui_hidden('cmd', 'add');
#	print &ui_submit("Добавить новую полку");
#	print &ui_form_end();
#	print "</td>";
#	print "</tr>";

	if (has_access("shelf_problem")) {
		my $pref = shelf_problems();
		if (scalar($pref) > 0) {
			my $url = "?cmd=problem";
			my @row = ("<a href=\"$url\">".$text{"shelf_problem_link"}."</a>");
			print &ui_columns_row(\@row);
		}
	}
	print &ui_table_end();
#		my %data = %{$store{'addr'}{'10.0.0.101'}};
#		print "<pre>".Dumper(\%data)."</pre>";

#	print "<pre>".Dumper(\%store)."</pre>";
#	print "<pre>".Dumper($store{$host}{'pool'})."</pre>";
#	print Dumper(\%store_cfg);
}


# FIXME: in progress
sub ui_shelf_add
{

	print &ui_form_start('shelf.cgi', 'post');
	print &ui_hidden('cmd', 'add');
	print &ui_table_start("Параметры новой полки", "width=100%", 2);
	print &ui_table_row("Имя полки", &ui_textbox("name", undef, 48));
	print &ui_table_row("IP-адрес", &ui_textbox("addr", undef, 48));
	print &ui_table_row("", &ui_submit("Зарегистрировать"));
	print &ui_table_end();
	print &ui_form_end();
}


sub ui_smart_row
{
	my ($name, $value) = @_;
	my $ret;
	my @tags = ();

	if (defined($value) && ($value > 0)) {
		my $style = "style=\"color: red; font-size: 1.1em\"";
		@tags = ($style, $style);
	}

	$ret = ui_table_row($name, $value, 1, \@tags);

	return $ret;
}


sub ui_vdev_smart
{
	my ($smart) = @_;
	my $ret;

	if (defined($smart->{'egdl'})) {
		# number of relocatable sectors in the list
		$ret .= ui_smart_row($text{'vdev_egdl'}, $smart->{'egdl'});
	}
	if (defined($smart->{'read-total-unc-errors'})) {
		# read errors
		$ret .= ui_smart_row($text{'vdev_read'}, $smart->{'read-total-unc-errors'});
	}
	if (defined($smart->{'write-total-unc-errors'})) {
		# write errors
		$ret .= ui_smart_row($text{'vdev_write'}, $smart->{'write-total-unc-errors'});
	}
	if (defined($smart->{'verify-total-unc-errors'})) {
		# verify errors
		$ret .= ui_smart_row($text{'vdev_verify'}, $smart->{'verify-total-unc-errors'});
	}
	if (defined($smart->{'relocated-sectors-count'})) {
		$ret .= ui_smart_row($text{'vdev_relsect'}, $smart->{'relocated-sectors-count'});
	}
	if (defined($smart->{'ssd-program-fail-count'})) {
		$ret .= ui_smart_row($text{'vdev_pfail'}, $smart->{'ssd-program-fail-count'});
	}
	if (defined($smart->{'ssd-erase-fail-count'})) {
		$ret .= ui_smart_row($text{'vdev_efail'}, $smart->{'ssd-erase-fail-count'});
	}
	if (defined($smart->{'end-to-end-error'})) {
		$ret .= ui_smart_row($text{'vdev_ete'}, $smart->{'end-to-end-error'});
	}
	if (defined($smart->{'reported-uncorrectable-errors'})) {
		$ret .= ui_smart_row($text{'vdev_uncorrected'}, $smart->{'reported-uncorrectable-errors'});
	}
	if (defined($smart->{'udma-crc-error-count'})) {
		$ret .= ui_smart_row($text{'vdev_crc'}, $smart->{'udma-crc-error-count'});
	}

	return $ret;
}


sub ui_vdev_info
{
	my ($vdev, $shelf) = @_;
	my $vptr = \%{$shelf->{'vdev'}{$vdev}};
	my $pool = vdev_is_used($vdev, \%{$shelf->{'pool'}});
	my $smart = \%{$vptr->{'smart'}};
	my $rrate = $smart->{'rotation_rate'};
	my $wdt = 6;
	my $ret;

	$ret .= ui_table_start(undef, undef, $wdt);
	$ret .= ui_table_row($text{'vdev_vid'}, $vptr->{'vid'});
	$ret .= ui_table_row($text{'vdev_pid'}, $vptr->{'pid'});
	$ret .= ui_table_row($text{'vdev_serial'}, $vptr->{'serial'});
	if (defined($vptr->{'revision'})) {
		$ret .= ui_table_row($text{'vdev_revision'}, $vptr->{'revision'});
	}
	if (defined($vptr->{'firmware'})) {
		$ret .= ui_table_row($text{'vdev_firmware'}, $vptr->{'firmware'});
	}
	$ret .= ui_table_row($text{'vdev_capacity'}, nice_size($vptr->{'capacity'}));
#	$ret .= ui_table_row($text{'vdev_phys_bsize'}, $vptr->{'physical'});
	$ret .= ui_table_row($text{'vdev_bsize'}, $vptr->{'logical'});
#	if ($vptr->{'ssd'} eq 1) {
#		$ret .= ui_table_row($text{'vdev_type'}, $text{'vdev_ssd'});
#	} else {
#		$ret .= ui_table_row($text{'vdev_type'}, $text{'vdev_rotational'});
#		$ret .= ui_table_row($text{'vdev_rrate'}, $rrate);
#	}

#	if (defined($rrate) && ($rrate =~ /^[0-9]+$/) && ($rrate > 0)) {
#		# sas/sata
#		$ret .= ui_table_row($text{'vdev_rrate'}, $rrate);
#		# number of relocatable sectors in the list
#		$ret .= ui_smart_row($text{'vdev_egdl'}, $vptr->{'egdl'});
#		# read errors
#		$ret .= ui_smart_row($text{'vdev_read'}, $vptr->{'read'});
#		# write errors
#		$ret .= ui_smart_row($text{'vdev_write'}, $vptr->{'write'});
#		# verify errors
#		$ret .= ui_smart_row($text{'vdev_verify'}, $vptr->{'verify'});
#	} else {
#		# SSD
#		$ret .= ui_table_row($text{'vdev_type'}, "SSD");
#
#		$ret .= ui_smart_row($text{'vdev_relsect'}, $vptr->{'relocated_sectors'});
#		# This attribute is used to count the number of NAND program fails
#		$ret .= ui_smart_row($text{'vdev_pfail'}, $vptr->{'prog_fail'});
#		# This attribute is used to count the number of NAND erase fails over the life of the drive
#		$ret .= ui_smart_row($text{'vdev_efail'}, $vptr->{'erase_fail'});
#		$ret .= ui_smart_row($text{'vdev_ete'}, $vptr->{'ete_err'});
#		$ret .= ui_smart_row($text{'vdev_uncorrected'}, $vptr->{'uncorrect'});
#		$ret .= ui_smart_row($text{'vdev_crc'}, $vptr->{'crc_error'});
#	}

#	my $status;
#	if (defined($pool) && ($pool ne "")) {
#		$status = $text{'vdev_in_use'}.", ".$shelf->{'pool'}{$pool}{'vdev'}{$vdev};
#	} else {
#		$status = $text{'vdev_not_used'};
#	}
#	$ret .= ui_table_row($text{'vdev_status'}, $status);
	$ret .= ui_table_row("vdev", $vdev);

	if (defined($vptr->{'enclosure'}) && ($vptr->{'enclosure'} >= 0)) {
		$ret .= ui_table_row($text{'vdev_enclosure'}, $vptr->{'enclosure'});
	}
	if (defined($vptr->{'slot'}) && ($vptr->{'slot'} >= 0)) {
		$ret .= ui_table_row($text{'vdev_slot'}, $vptr->{'slot'});
	}

#	$ret .= ui_columns_header(["S.M.A.R.T"], ["colspan=$wdt align=\"center\""]);
	$ret .= ui_vdev_smart($smart);

	$ret .= ui_table_end();

	return $ret;
}


sub ui_show_vdev
{
	my ($shelf_name, $dev, $shelf, $code) = @_;
	my $spool = \%{$shelf->{'pool'}};
	my $vdev = \%{$shelf->{'vdev'}{$dev}};
	my $smart = \%{$vdev->{'smart'}};
	my $pool = vdev_is_used($dev, $spool);
	my $title = $dev.", ".nice_size($vdev->{'capacity'});
	my $status;
	my $out;
	my $info;

	if (defined($pool) && ($pool ne "")) {
		$status = $spool->{$pool}{'vdev'}{$dev};
		if ($status =~ /online/i) {
			$status = "<font color=\"".$pool_color{'ok'}."\">". $status . "</font>";
		} elsif ($status =~ /degraded/i) {
			$status = "<font color=\"blue\">". $status . "</font>";
		} elsif ($status =~ /faulted/i) {
			$status = "<font color=\"".$pool_color{'crit'}."\">". $status . "</font>";
		}
	} else {
		$status = "<font color=\"grey\">".$text{'vdev_not_used'}."</font>";
	}
	$title .= " (".$status.")";
	if ($code > 0) {
		my $errors;
		# SAS/SATA
		if ($code & 0x002) {
			# egdl
			$errors .= $text{'err_egdl'}.": ".$smart->{'egdl'};
		}
		if ($code & 0x004) {
			# read
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_rd'}.": ".$smart->{'read'};
		}
		if ($code & 0x008) {
			# write
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_wr'}.": ".$smart->{'write'};
		}
		if ($code & 0x010) {
			# verify
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_vf'}.": ".$smart->{'verify'};
		}
		# SSD
		if ($code & 0x020) {
			# relocated_sectors
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_rel'}.": ". $smart->{'relocated-sectors-count'};
		}
		if ($code & 0x040) {
			# prog_fail
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_pf'}.": ". $smart->{'ssd-program-fail-count'};
		}
		if ($code & 0x080) {
			# erase_fail
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_ef'}.": ".$smart->{'ssd-erase-fail-count'};
		}
		if ($code & 0x100) {
			# ete_err
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_ete'}.": ".$smart->{'end-to-end-error'};
		}
		if ($code & 0x200) {
			# uncorrect
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_unc'}.": ". $smart->{'reported-uncorrectable-errors'};
		}
		if ($code & 0x400) {
			# crc_error
			if (defined($errors)) {
				$errors .= "; ";
			}
			$errors .= $text{'err_crc'}.": ".$smart->{'udma-crc-error-count'};
		}
		$title .= "; <font color=\"red\">$errors</font>";
	}

	$info = ui_vdev_info($dev, $shelf);

	$out = ui_hidden_table_start($title, undef, 1, $dev, 0);
	$out .= ui_columns_row([$info]);
	$out .= ui_hidden_table_end($dev);

	return $out;
}


sub ui_show_shelf
{
	my ($addr, $pref) = @_;
	my $shelf = \%{$store{'shelf'}{$addr}};
	my $name = $shelf->{'name'};
	my $code = $pref->{$addr}{'code'};
	my $vdevs = \%{$shelf->{'vdev'}};
	my $style = "style=\"padding-left: 1em; padding-right: 1em\"";
	my $ret;
	my $out;

	if ($code & 1) {
		# There are problems in the pool(s)
	}
	foreach my $d (sort keys %{$vdevs}) {
		my $vdev = $vdevs->{$d};
		$code = vdev_has_problem($vdev);
		if ($code eq 0) {
			next;
		}
		$out .= ui_show_vdev($name, $d, $shelf, $code);
	}
	$ret .= ui_hidden_table_start($name, undef, 3, $name, 1);
	$ret .= ui_columns_row([$out]);
	$ret .= ui_hidden_table_end($name);

	return $ret;
}


sub ui_shelf_detail
{
	my ($shelf_name) = @_;
	my $addr = $store{'config'}{$shelf_name};
	my $shelf = \%{$store{'shelf'}{$addr}};
	my $vdevs = \%{$shelf->{'vdev'}};
	my $out;

	print "<h1>".text('shelf_hdr_detail', $shelf_name)."</h1>";
	$out .= ui_table_start($shelf_name, "width=\"100%\"", 1);
	foreach my $d (sort keys %{$vdevs}) {
		$out .= ui_show_vdev($shelf_name, $d, $shelf, 0);
	}
	$out .= ui_table_end();

	print $out;
	shelf_footer();
#	print "<pre>".Dumper(\%store)."</pre>";
}


sub ui_shelf_problem
{
	my $pref = shelf_problems();
	my $out;

	if (scalar($pref) == 0) {
		print "<h2>".$text{'shelf_no_problems"'}."</h2>";
		return;
	}

	print "<h1>".$text{'shelf_hdr_problem'}."</h1>";
	$out .= ui_table_start(undef, "width=\"100%\"", 1);
	foreach my $sh (sort keys %{$pref}) {
		$out .= ui_show_shelf($sh, $pref);
	}
	$out .= ui_table_end();

	print $out;
	shelf_footer();
#	print "<pre>".Dumper(\%store)."</pre>";
}


### Main ###

my $cmd;

if (defined($in{'cmd'})) {
	$cmd = $in{'cmd'};
}

#if ((!defined $cmd) || ($cmd eq "")) {
#	$refresh = "<meta http-equiv=\"refresh\" content=\"60\">";
#}

my $cgi = "recollect.cgi";
if (defined($cmd)) {
	$cgi .= "?cmd=".$cmd;
	if ($cmd eq "detail") {
		my $shelf = $in{'shelf'};
		$cgi .= "&shelf=".$shelf;
	}
}
my $link = "<a href=\"".$cgi."\">".$text{'shelf_recollect'}."</a>";

ui_print_header(undef, $text{'index_title'}, undef, undef, 0, 1, 0,
        $link, $refresh, undef, $text{'index_version'} );

if (!defined($cmd) || ($cmd eq "")) {
	goto SKIP;
} elsif ($cmd eq "add") {
	ui_shelf_add();
} elsif ($cmd eq "detail") {
	my $shelf = trim($in{'shelf'});
	if (!defined($shelf) || ($shelf eq "")) {
		goto SKIP;
	}
	ui_shelf_detail($shelf);
} elsif ($cmd eq "problem") {
	ui_shelf_problem();
} else {
	# Show all shelves
	SKIP:
	shelves_show();
}

ui_print_footer("/", $text{'index'});

### END main ###
