BEGIN { push(@INC, ".."); };
use WebminCore;
use POSIX qw(strftime);
use Text::Trim qw(trim);
use Data::Dumper;
use Sys::Hostname;
init_config();

my %access = &get_module_acl();
my $ping_cmd = "/usr/sbin/ping";
#my $iscsiadm_cmd = "/usr/sbin/iscsiadm";
#my $iostat_cmd = "/usr/bin/iostat";
#my $kstat_cmd = "/usr/bin/kstat";
my $consul_cmd = "/usr/bin/consul";
my $cfg_file = "/etc/default/store";
my $devpath = "/dev/rdsk/";
my $devpatt = qr/$devpath/;

#my $fake_path = "fake";
my $cache_path = "/var/store/shelfctl";

our $err;
our %store;
our $host = hostname;
our %pool_color = (
	'ok' => "#00D000",
	'warn' => "yellow",
	'info' => "#BBBBBB",
	'crit' => "red"
);

$store{'config'} = store_config_load(\%{$store{'shelf'}});
iscsi_list_target(\%{$store{'target'}}, \%{$store{'vdev'}}, \%{$store{'sd'}});


sub has_access
{
	my ($acc) = @_;

	return $access{$acc};
}


sub ping
{
	my ($addr) = @_;
	my @res = &backquote_command("${ping_cmd} $addr 20");

	return $?;
}


sub parse_cache_line
{
	my ($cfg, $href, $dref) = @_;
	my $name = $dref->[0];
	my $addr = $cfg->{$name};

	if (!defined($addr) || ($addr eq "")) {
		$href = \%store;
		$addr = $host;
	}
	if ($dref->[1] =~ /uptime/) {
		$href->{$addr}{'uptime'} = $dref->[2];
	} elsif ($dref->[1] =~ /disk/) {
		if ($dref->[3] =~ /smart/) {
			my @arg = split /:/, $dref->[4];
			$href->{$addr}{'vdev'}{$dref->[2]}{'smart'}{$arg[0]} = $arg[1];
		} else {
			my @arg = split /:/, $dref->[3];
			$href->{$addr}{'vdev'}{$dref->[2]}{$arg[0]} = $arg[1];
		}
	} elsif ($dref->[1] =~ /pool/) {
		my $name = $dref->[2];
		if ($dref->[3] =~ /vdev/) {
			my @arg = split /:/, $dref->[4];
			$href->{$addr}{'pool'}{$name}{'vdev'}{$arg[0]} = $arg[1];
		} else {
			my @arg = split /:/, $dref->[3];
			$href->{$addr}{'pool'}{$name}{$arg[0]} = $arg[1];
		}
	}
}


# Loads data of the shelves from a cache (file of consul)
sub cache_load
{
	my ($cfg, $href) = @_;
	my $cache = $cache_path."/shelf.cache";

	open(CFG, $cache) || return ;
	flock(CFG, 1);
	while (my $line=<CFG>) {
		chomp $line;
		my @data = split /\//, $line;
		parse_cache_line($cfg, $href, \@data);
	}
	close(CFG);
}


sub store_config_load
{
	my ($href) = @_;
	my $cmd = "/bin/sh -c '. $cfg_file;set'";
	my @res = &backquote_command($cmd);
	my %hash;

	chomp @res;

	foreach my $line (@res) {
		my $el = trim($line);
		if ($el eq "") {
			next;
		}
		(my $word1, my $word2) = split /=/, $el;
		$word2 =~ s/^'(.*)'$/$1/;
		if (($el =~ /^pool_.*_targets.*=/i) && !($el =~ /^pool_.*_targets_min.*=/i)) {
			if (($word2 eq "") || ($word2 eq 0)) {
				$word2 = undef;
			}
			my @tlist = split / /, $word2;
			$hash{$word1} = \@tlist;
			next;
		} elsif ($el =~ /^shelf_.*=/i) {
			if (($word2 eq "") || ($word2 eq 0)) {
				$word2 = undef;
			}
			$word1 =~ s/shelf_//i;
			$hash{$word1} = $word2;
			$$href{$word2}{'name'} = $word1;
		} elsif ($el =~ /^pool_/i) {
			if (($word2 eq "") || ($word2 eq 0)) {
				$word2 = undef;
			}
			$hash{$word1} = $word2;
		}
	}

	iscsi_list_address($href);
	cache_load(\%hash, $href);

	return \%hash;
}


sub store_config_save
{
	my (%hash) = @_;

	open(F, "+< $cfg_file") || die;
	flock(F, 2);

	# FIXME: переделать на два файла!
	foreach $key (keys %hash) {
		if ($key =~ /^STORE_TARGETS_/) {
			printf F "$key=\"%s\"\n", join(" ", @{$hash{$key}});
		} elsif ($key =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
			my $str = "ip".$key;
			$str =~ s/\./_/g;
			printf F "$str=\"%s\"\n", join(" ", @{$hash{$key}});
		} else {
			printf F "$key=\"%s\"\n", $hash{$key};
		}
	}
	close F;
}


sub smart_prediction
{
	my ($href) = @_;
	my %smart = %{$href};
	my $ret = 0;
	my $n = 0;
	my $N = 0;

	foreach my $vdev (sort keys %smart) {
		my %dsk = %{$smart{$vdev}};
		my $rrate = $dsk{'rotation_rate'};
		if ($rrate eq "SSD") {
			# SSD
			$N++;
			if (($dsk{'relocated'} > 0) || ($dsk{'uncorrect'} > 0) ||
				($dsk{'prog_fail'} > 0) || ($dsk{'crc_error'} > 0) ||
				($dsk{'ete_errors'} > 0) || ($dsk{'erase_fail'} > 0)) {
				$n++;
			}
		} elsif (($rrate ne "") && ($rrate > 0)) {
			# SAS/SATA
			$N++;
			if (($dsk{'defect_list'} > 0) || ($dsk{'read'} > 0) ||
				($dsk{'write'} > 0) || ($dsk{'verify'} > 0)) {
				$n++;
			}
		}
	}
	if ($N > 0) {
		$ret = ($n / $N) * 100.0;
	} else {
		$ret = 0.0;
	}

	return $ret;
}


sub iscsi_list_address
{
	my ($href) = @_;
	my $cmd;
	my @res;
	my $ip, $name;
	my $cache = $cache_path."/list_discovery-address.cache";

	open(CFG, $cache) || return ;
	flock(CFG, 1);
	while (my $line=<CFG>) {
		chomp $line;
		if ($line =~ /iSCSI Initiator Service is disabled/) {
			last;
		}
		my @arg = split /: /, trim($line);
		chomp @arg;
		if ($arg[0] =~ /Discovery Address/) {
			my @val = split /:/, $arg[1];
			chomp(@val);
			$ip = $val[0];
		} elsif ($arg[0] =~ /Target name/) {
			$name = $arg[1];
			$store{'target'}{$name} = {};
			push(@{$$href{$ip}{'target'}}, $name);
		}
	}
	close(CFG);
}


sub iscsi_list_target
{
	my ($tref, $vref, $sref) = @_;
	my $cmd;
	my @res;
	my $target;
	my $lun;
	my %vdev = ();
	my $cache = $cache_path."/list_target.cache";

	open(CFG, $cache) || return ;
	flock(CFG, 1);
	my $i = 0;
	while (my $line=<CFG>) {
		chomp $line;
		my @arg = split /: /, trim($line);
		chomp @arg;
		if ($arg[0] =~ /Target/) {
			$target = $arg[1];
		} elsif ($arg[0] =~ /Alias/) {
			$$tref{$target}{'alias'} = $arg[1];
		} elsif ($arg[0] =~ /TPGT/) {
			$$tref{$target}{'tpgt'} = $arg[1];
		} elsif ($arg[0] =~ /ISID/) {
			$$tref{$target}{'isid'} = $arg[1];
		} elsif ($arg[0] =~ /Connections/) {
			$$tref{$target}{'connections'} = $arg[1];
		} elsif ($arg[0] =~ /CID/) {
			$$tref{$target}{'cid'} = $arg[1];
		} elsif ($arg[0] =~ /IP address \(Local\)/) {
			my @val = split /:/, $arg[1];
			chomp(@val);
			$$tref{$target}{'ip'} = $val[0];
			$$tref{$target}{'sport'} = $val[1];
		} elsif ($arg[0] =~ /IP address \(Peer\)/) {
			my @val = split /:/, $arg[1];
			chomp(@val);
			$$tref{$target}{'peer'} = $val[0];
			$$tref{$target}{'dport'} = $val[1];
		} elsif ($arg[0] =~ /Discovery Method/) {
			$$tref{$target}{'discovery_method'} = $arg[1];
		} elsif ($arg[0] =~ /Data Sequence In Order/) {
			$$tref{$target}{'seq_in_order'} = $arg[1];
		} elsif ($arg[0] =~ /Data PDU In Order/) {
			$$tref{$target}{'pdu_in_order'} = $arg[1];
		} elsif ($arg[0] =~ /Default Time To Retain/) {
			$$tref{$target}{'retain'} = $arg[1];
		} elsif ($arg[0] =~ /Default Time To Wait/) {
			$$tref{$target}{'wait'} = $arg[1];
		} elsif ($arg[0] =~ /Error Recovery Level/) {
			$$tref{$target}{'r_level'} = $arg[1];
		} elsif ($arg[0] =~ /First Burst Length/) {
			$$tref{$target}{'fb_len'} = $arg[1];
		} elsif ($arg[0] =~ /Immediate Data/) {
			$$tref{$target}{'imm_data'} = $arg[1];
		} elsif ($arg[0] =~ /Initial Ready To Transfer (R2T)/) {
			$$tref{$target}{'r2t'} = $arg[1];
		} elsif ($arg[0] =~ /Max Burst Length/) {
			$$tref{$target}{'mbl'} = $arg[1];
		} elsif ($arg[0] =~ /Max Outstanding R2T/) {
			$$tref{$target}{'max_out_r2t'} = $arg[1];
		} elsif ($arg[0] =~ /Max Receive Data Segment Length/) {
			$$tref{$target}{'max_seg_length'} = $arg[1];
		} elsif ($arg[0] =~ /Max Connections/) {
			$$tref{$target}{'max_connections'} = $arg[1];
		} elsif ($arg[0] =~ /Header Digest/) {
			$$tref{$target}{'hdr_digest'} = $arg[1];
		} elsif ($arg[0] =~ /Data Digest/) {
			$$tref{$target}{'data_digest'} = $arg[1];
		} elsif ($arg[0] =~ /LUN/) {
			$lun = $arg[1];
			if (!defined($$tref{$target}{'luns'})) {
				$$tref{$target}{'luns'} = ();
			}
		} elsif ($arg[0] =~ /Vendor/) {
			$vdev{'vid'} = $arg[1];
		} elsif ($arg[0] =~ /Product/) {
			$vdev{'pid'} = $arg[1];
		} elsif ($arg[0] =~ /OS Device Name/) {
			$vdev{'device'} = $arg[1];
			my $dev = $arg[1];
			$dev =~ s/$devpatt//g;
			$dev =~ s/s2$//g;
			$i++;
			$vdev{'sd'} = $store{$host}{'vdev'}{$dev}{'sd'};
			$vdev{'size'} = $store{$host}{'vdev'}{$dev}{'capacity'};
			$$tref{$target}{'luns'}{$lun}{'vdev'} = $dev;
			$$tref{$target}{'luns'}{$lun}{'sd'} = $vdev{'sd'};
			my %vd = %vdev;
			$$vref{$dev} = \%vd;
			my $sd = $vdev{'sd'};
			$$sref{$sd} = $dev;
			$vdev = ();
		}
	}
	close(CFG);
}


sub iscsi_add_addr
{
	my ($addr) = @_;
	my $cmd = "${iscsiadm_cmd} add discovery-address $addr";
	my @res = &backquote_command($cmd);

	if ($? eq 0) {
		push(@{$store_cfg{'STORE_TARGETS_1'}}, $addr);
	}

	return ($?);
}


# Returns size exported by a shelf
#
# $addr		- IP-address of a shelf
sub shelf_target_size
{
	my ($addr) = @_;
	my $shelf = \%{$store{'shelf'}{$addr}};
	my $target = $shelf->{'target'};
	my $ret = 0;

	if (!defined($target)) {
		return 0;
	}

	my @vdev = ();

	foreach my $t (@{$target}) {
		my $luns = $store{'target'}{$t}{'luns'};
		foreach my $l (keys %{$luns}) {
			my $d = $luns->{$l}{'vdev'};
			push (@vdev, $d) unless grep{$_ eq $d} @vdev;
		}
	}

	foreach my $d (@vdev) {
		$ret += $store{'vdev'}{$d}{'size'};
	}

	return $ret;
}


sub shelf_scan_pools
{
	my ($ref) = @_;
	my $code = 0;

	foreach my $p (sort keys %{$ref}) {
		my $pref = \%{$ref->{$p}};
		if ($pref->{'status'} =~ /degraded/i) {
			$code |= 0x01;
			last;
		}
	}

	return $code;
}


sub vdev_has_problem
{
	my ($vdev, $ref) = @_;
	my $smart = \%{$vdev->{'smart'}};
	my $code = 0;

	if ($smart->{'egdl'} > 0) {
		$code |= 0x02;
	}
	if ($smart->{'read'} > 0) {
		$code |= 0x04;
	}
	if ($smart->{'write'} > 0) {
		$code |= 0x08;
	}
	if ($smart->{'verify'} > 0) {
		$code |= 0x10;
	}
	if ($smart->{'relocated-sectors-count'} > 0) {
		$code |= 0x20;
	}
	if ($smart->{'ssd-program-fail-count'} > 0) {
		$code |= 0x40;
	}
	if ($smart->{'ssd-erase-fail-count'} > 0) {
		$code |= 0x80;
	}
	if ($smart->{'end-to-end-error'} > 0) {
		$code |= 0x100;
	}
	if ($smart->{'reported-uncorrectable-errors'} > 0) {
		$code |= 0x200;
	}
	if ($smart->{'udma-crc-error-count'} > 0) {
		$code |= 0x400;
	}

	return $code;
}


# Returns pool name which contains the vdev
#
sub vdev_is_used
{
	my ($vdev, $ref) = @_;
	my $ret;

	foreach $p (sort keys %{$ref}) {
		if (defined($ref->{$p}{'vdev'}{$vdev})) {
			$ret = $p;
			last;
		}
	}

	return $ret;
}


sub shelf_scan_vdevs
{
	my ($ref) = @_;
	my $code = 0;

	foreach my $d (sort keys %{$ref}) {
		my $vdev = \%{$ref->{$d}};
		$code = vdev_has_problem($vdev);
		if ($code > 0) {
			last;
		}
	}

	return $code;
}


# Returns hash with shelf IP and bitwise code of the problems
sub shelf_problems
{
	my %hash = ();

	foreach my $s (sort keys %{$store{'shelf'}}) {
		my $ref = \%{$store{'shelf'}{$s}};
		my $code = 0;
		if (defined($ref->{'pool'})) {
			$code |= shelf_scan_pools(\%{$ref->{'pool'}});
		}
		if (defined($ref->{'vdev'})) {
			$code |= shelf_scan_vdevs(\%{$ref->{'vdev'}});
		}
		if ($code > 0) {
			$hash{$s}{'code'} = $code;
		}
	}

	return \%hash;
}


# Returns new pool status
#
# $cs		- current status
# $ns		- new (comparise) status
sub pool_get_status
{
	my ($cs, $ns) = @_;
	my $ret = $cs;

	if ($cs =~ /online/i) {
		if ($ns =~ /online/i) {
			$ret = $cs;
		} else {
			$ret = $ns;
		}
	} elsif ($cs =~ /degraded/i) {
		if (($ns =~ /faulted/i) || ($ns =~ /unavail/i)) {
			$ret = $ns;
		} else {
			$ret = $cs;
		}
	}

	return $ret;
}


# Returns status of the shelf
#
# $stgt		- shelf's targets
# $info		- out array
sub shelf_info
{
	my ($info, $stgt) = @_;
	my $target = $store{'target'}; #{'luns'}{'0'}{'vdev'}
	my $pools = \%{$store{$host}{'pool'}};
	my %vdevs = ();
	my $shelf_pool = "";

	for my $t (@{$stgt}) {
		my $luns = \%{$target->{$t}{'luns'}};
		my $peer = $target->{$t}{'peer'};
		my $sh_pool = \%{$store{'shelf'}{$peer}{'pool'}};
		for my $p (keys %{$sh_pool}) {
			my $cps = $sh_pool->{$p}{'status'};
#			print STDERR Dumper($cps);
			if (!defined($shelf_pool) || ($shelf_pool eq "")) {
				$shelf_pool = $cps;
				next;
			} else {
				$shelf_pool = pool_get_status($shelf_pool, $cps);
			}
		}
#		print STDERR Dumper($luns);
		for my $lu (keys %{$luns}) {
			my $dev = $luns->{$lu}{'vdev'};
			$vdevs{$dev} = 1;
		}
	}

	for my $p (keys %{$pools}) {
		my $vdev = \%{$pools->{$p}{'vdev'}};
		for my $vd (keys %vdevs) {
			if (defined($vdev->{$vd})) {
				$info->{'status'} = $vdev->{$vd};
				goto OUT;
			}
		}
	}
OUT:
	if (defined($info->{'status'})) {
#		print STDERR Dumper($info);
#		print STDERR Dumper($shelf_pool);
		$info->{'status'} = pool_get_status($info->{'status'}, $shelf_pool);
		my $status = $info->{'status'};
		if ($status =~ /online/i ) {
			$info->{'color'} = $pool_color{'ok'};
		} elsif	(($status =~ /resilvering/i) || ($status =~ /repairing/i) || ($status =~ /offline/i)) {
			$info->{'color'} = $pool_color{'warn'};
		} elsif	(($status =~ /degraded/i)) {
			$info->{'color'} = $pool_color{'info'};
		} elsif	(($status =~ /faulted/i) || ($status =~ /unavail/i)) {
			$info->{'color'} = $pool_color{'crit'};
		}
	} elsif (!defined($info->{'uptime'})) {
		$info->{'status'} = "UNAVAIL";
		$info->{'color'} = 'red';
	}
#	print STDERR Dumper($target);
#	$info->{'status'} = 'status';
}


sub do_shelf_add
{
	my ($name, $addr) = @_;

	if (ping($addr) eq 0) {
		# добавляем
		my @list = iscsi_list_address();
#		if ($addr ~~ @list) {
#			print "Уже есть такой";
#		} else {
#			iscsi_add_addr($addr);
#			$store_cfg{$addr} = $name;
#			store_config_save(%store_cfg);
#			@list = iscsi_list_address();
#		}
	}
}


sub consul_load
{
	my %hash;
	my @res = &backquote_command("${consul_cmd} kv get -recurse shelf");

	chomp(@res);
	foreach $line (@res) {
		my @arg = split /\//, $line;
		chomp @arg;
		if ($arg[3] =~ ":" ){
			my $str = $arg[3];
			($arg[3], $arg[4]) = split /:/, $str;
		}
		$hash{$arg[1]}{$arg[2]}{$arg[3]} = $arg[4];
	}

	return (\%hash);
}

1;
